/*-
 * Copyright 2017 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <locale.h>

#include "vcl.h"
#include "cache/cache.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"

#include "vcc_if.h"

#define PCRE2_CODE_UNIT_WIDTH 8
#include "pcre2.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod pcre2 error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod pcre2 error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
        VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
        ERR((ctx), msg ", out of space")

#define INIT(ctx) (((ctx)->method & VCL_MET_INIT) != 0)

#define SET_CTX_PARAM(ctx, param, type)	do {				\
		if ((param) != 0) AZ(pcre2_set_##param((ctx), (type)(param))); \
	} while(0)

#define CHECK_UINT32_RANGE(limit, vcl_name, caller, ret)		\
	do {								\
		if ((limit) < 0 || (limit) > UINT32_MAX) {		\
			VERR(ctx, #limit " (%lld) out of range in %s"	\
			     caller " (must be >= 0 and <= %" PRIu32 ")", \
			     (long long)(limit), (vcl_name), UINT32_MAX); \
			return ret;					\
		}							\
	} while(0)

struct vmod_pcre2_regex {
	unsigned	magic;
#define VMOD_PCRE2_REGEX_MAGIC 0x3adb2a78
	pcre2_code	*code;
	char		*vcl_name;
};

/* PRIV_CALL scope for match */
struct match_call {
	unsigned		magic;
#define VMOD_PCRE2_MATCH_CALL_MAGIC 0x60e5bd33
	pcre2_match_context	*mctx;
	uint32_t		options;
};

static int have_jit = 0;

void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		VSB_vprintf(ctx->msg, fmt, args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
	else if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		/* Should this ever happen in vcl_fini() ... */
		VSL(SLT_VCL_Error, 0, fmt, args);
	va_end(args);
}

static void
report_pcre2_err(VRT_CTX, int errcode, const char * const restrict msg,
		 const char * const restrict post)
{
	int ret;

	uintptr_t snap = WS_Snapshot(ctx->ws);
	unsigned bytes = WS_Reserve(ctx->ws, 0);
	char *buf = WS_Front(ctx->ws);

	ret = pcre2_get_error_message(errcode, (PCRE2_UCHAR *)buf,
				      (PCRE2_SIZE)bytes);
	if (ret == PCRE2_ERROR_BADDATA) {
		WS_ReleaseP(ctx->ws, buf);
		VERR(ctx, "%s (unknown error code)%s", msg, post);
	}
	else if (ret == PCRE2_ERROR_NOMEMORY) {
		WS_Release(ctx->ws, strlen(buf) + 1);
		VERR(ctx, "%s: %s (truncated error message)%s", msg, buf,
		     post);
	}
	else {
		WS_Release(ctx->ws, ret);
		VERR(ctx, "%s: %s%s", msg, buf, post);
	}
	WS_Reset(ctx->ws, snap);
}

static inline void
set_opt(uint32_t *options, VCL_BOOL vmod_opt, uint32_t pcre2_opt)
{
	if (vmod_opt)
		*options |= pcre2_opt;
}

/* Event function */

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) ctx;
	(void) priv;

	if (e == VCL_EVENT_LOAD) {
		int ret;
		uint32_t where;

		/*
		 * A VCL_INT must never be larger than PCRE2_SIZE_MAX, so
		 * that some assignments from INT params cannot be out of
		 * range.
		 * XXX: C11 static_assert would catch this at compile
		 * time.
		 */
		assert(sizeof(VCL_INT) <= sizeof(PCRE2_SIZE));

		ret = pcre2_config(PCRE2_CONFIG_JIT, &where);
		assert(ret >= 0);
		have_jit = where;
	}
	return 0;
}

/* Object regex */

static void *
ws_malloc(PCRE2_SIZE sz, void *ws)
{
	return WS_Alloc(ws, sz);
}

static void
ws_free(void *ptr, void *ws)
{
	WS_Assert_Allocated(ws, ptr, 0);
}

static void
match_call_free(void *priv)
{
	struct match_call *mctx_call;

	if (priv == NULL)
		return;
	CAST_OBJ(mctx_call, priv, VMOD_PCRE2_MATCH_CALL_MAGIC);
	pcre2_match_context_free(mctx_call->mctx);
	FREE_OBJ(mctx_call);
}

VCL_VOID
vmod_regex__init(VRT_CTX, struct vmod_pcre2_regex **regexp,
		 const char *vcl_name, VCL_STRING pattern,
		 VCL_BOOL allow_empty_class, VCL_BOOL anchored, VCL_ENUM bsrs,
		 VCL_BOOL alt_bsux, VCL_BOOL alt_circumflex,
		 VCL_BOOL alt_verbnames, VCL_BOOL caseless,
		 VCL_BOOL dollar_endonly, VCL_BOOL dotall, VCL_BOOL dupnames,
		 VCL_BOOL extended, VCL_BOOL firstline, VCL_STRING locale,
		 VCL_BOOL match_unset_backref, VCL_INT max_pattern_length,
		 VCL_BOOL multiline, VCL_BOOL never_backslash_c,
		 VCL_BOOL never_ucp, VCL_BOOL never_utf, VCL_ENUM newlines,
		 VCL_BOOL no_auto_capture, VCL_BOOL no_auto_possess,
		 VCL_BOOL no_dotstar_anchor, VCL_BOOL no_start_optimize,
		 VCL_BOOL no_utf_check, VCL_INT parens_nest_limit,
		 VCL_BOOL ucp, VCL_BOOL ungreedy, VCL_BOOL use_offset_limit,
		 VCL_BOOL utf)
{
	struct vmod_pcre2_regex *regex;
	pcre2_code *code;
	pcre2_compile_context *ccontext;
	uint32_t options = 0, val;
	int ret, err_code = 0;
	PCRE2_SIZE err_offset;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(regexp);
	AZ(*regexp);
	AN(vcl_name);
	if (pattern == NULL) {
		VERR(ctx, "pattern is NULL in %s constructor", vcl_name);
		return;
	}
	if ((ccontext = pcre2_compile_context_create(NULL)) == NULL) {
		VERR(ctx, "failed to create compile context in %s constructor",
		     vcl_name);
		return;
	}
	if (max_pattern_length < 0) {
		VERR(ctx, "max_pattern_len (%lld) must be >= 0 in %s "
		     "constructor", (long long)max_pattern_length, vcl_name);
		return;
	}
	CHECK_UINT32_RANGE(parens_nest_limit, vcl_name, " constructor",);

	/* XXX check illegal combinations such as never_ucp && ucp ... ? */
	set_opt(&options, anchored,		PCRE2_ANCHORED);
	set_opt(&options, allow_empty_class,	PCRE2_ALLOW_EMPTY_CLASS);
	set_opt(&options, alt_bsux,		PCRE2_ALT_BSUX);
	set_opt(&options, alt_circumflex,	PCRE2_ALT_CIRCUMFLEX);
	set_opt(&options, alt_verbnames,	PCRE2_ALT_VERBNAMES);
	set_opt(&options, caseless,		PCRE2_CASELESS);
	set_opt(&options, dollar_endonly,	PCRE2_DOLLAR_ENDONLY);
	set_opt(&options, dotall,		PCRE2_DOTALL);
	set_opt(&options, dupnames,		PCRE2_DUPNAMES);
	set_opt(&options, extended,		PCRE2_EXTENDED);
	set_opt(&options, firstline,		PCRE2_FIRSTLINE);
	set_opt(&options, match_unset_backref,	PCRE2_MATCH_UNSET_BACKREF);
	set_opt(&options, multiline,		PCRE2_MULTILINE);
	set_opt(&options, never_backslash_c,	PCRE2_NEVER_BACKSLASH_C);
	set_opt(&options, never_ucp,		PCRE2_NEVER_UCP);
	set_opt(&options, never_utf,		PCRE2_NEVER_UTF);
	set_opt(&options, no_auto_capture,	PCRE2_NO_AUTO_CAPTURE);
	set_opt(&options, no_auto_possess,	PCRE2_NO_AUTO_POSSESS);
	set_opt(&options, no_dotstar_anchor,	PCRE2_NO_DOTSTAR_ANCHOR);
	set_opt(&options, no_start_optimize,	PCRE2_NO_START_OPTIMIZE);
	set_opt(&options, no_utf_check,		PCRE2_NO_UTF_CHECK);
	set_opt(&options, ucp,			PCRE2_UCP);
	set_opt(&options, ungreedy,		PCRE2_UNGREEDY);
	set_opt(&options, use_offset_limit,	PCRE2_USE_OFFSET_LIMIT);
	set_opt(&options, utf,			PCRE2_UTF);

	SET_CTX_PARAM(ccontext, max_pattern_length, PCRE2_SIZE);
	SET_CTX_PARAM(ccontext, parens_nest_limit, uint32_t);
	if (bsrs != NULL) {
		if (strcmp("ANYCRLF", bsrs) == 0)
			val = PCRE2_BSR_ANYCRLF;
		else if (strcmp("UNICODE", bsrs) == 0)
			val = PCRE2_BSR_UNICODE;
		else
			WRONG("Illegal bsr enum value");
		AZ(pcre2_set_bsr(ccontext, val));
	}
	if (locale != NULL) {
		const unsigned char *tables;
		char *saved_lc = setlocale(LC_CTYPE, NULL);

		AN(saved_lc);
		if (setlocale(LC_CTYPE, locale) == NULL) {
			AN(setlocale(LC_CTYPE, saved_lc));
			VERR(ctx, "Cannot set locale %s in %s constructor",
			     locale, vcl_name);
			return;
		}
		tables = pcre2_maketables(NULL);
		if (tables == NULL) {
			AN(setlocale(LC_CTYPE, saved_lc));
			VERRNOMEM(ctx, "Creating tables for locale %s in "
				  "%s constructor", locale, vcl_name);
			return;
		}
		AZ(pcre2_set_character_tables(ccontext, tables));
		AN(setlocale(LC_CTYPE, saved_lc));
	}
	if (newlines != NULL) {
		if (strcmp("CR", newlines) == 0)
			val = PCRE2_NEWLINE_CR;
		else if (strcmp("LF", newlines) == 0)
			val = PCRE2_NEWLINE_LF;
		else if (strcmp("CRLF", newlines) == 0)
			val = PCRE2_NEWLINE_CRLF;
		else if (strcmp("ANYCRLF", newlines) == 0)
			val = PCRE2_NEWLINE_ANYCRLF;
		else if (strcmp("ANY", newlines) == 0)
			val = PCRE2_NEWLINE_ANY;
		else
			WRONG("Illegal newline enum value");
		AZ(pcre2_set_newline(ccontext, val));
	}

	/* XXX set the length via parameter */
	code = pcre2_compile((PCRE2_SPTR)pattern, PCRE2_ZERO_TERMINATED,
			     options, &err_code, &err_offset, ccontext);
	if (code == NULL) {
		char *msg, *offset_msg;
		uintptr_t snap = WS_Snapshot(ctx->ws);

		if ((msg = WS_Printf(ctx->ws,
				     "Cannot compile '%s' in %s constructor",
				     pattern, vcl_name)) == NULL)
			msg = "";
		if ((offset_msg = WS_Printf(ctx->ws, " at offset %zu",
					    err_offset)) == NULL)
			offset_msg = "";
		report_pcre2_err(ctx, err_code, msg, offset_msg);
		WS_Reset(ctx->ws, snap);
		pcre2_compile_context_free(ccontext);
		return;
	}
	if (have_jit) {
		/* XXX check option compatibility; disable via param */
		/* XXX set complete or soft/hard partial via param */
		options |= PCRE2_JIT_COMPLETE;
		ret = pcre2_jit_compile(code, options);
		if (ret != 0) {
			char *msg;
			uintptr_t snap = WS_Snapshot(ctx->ws);

			if ((msg = WS_Printf(ctx->ws, "Cannot jit-compile '%s' "
					     "in %s constructor", pattern,
					     vcl_name)) == NULL)
				msg = "";
			report_pcre2_err(ctx, ret, msg, "");
			WS_Reset(ctx->ws, snap);
			pcre2_compile_context_free(ccontext);
			return;
		}
	}

	ALLOC_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	AN(regex);
	*regexp = regex;
	regex->code = code;
	regex->vcl_name = strdup(vcl_name);
	AN(regex->vcl_name);
	pcre2_compile_context_free(ccontext);
}

VCL_VOID
vmod_regex__fini(struct vmod_pcre2_regex **regexp)
{
	struct vmod_pcre2_regex *regex;

	if (regexp == NULL)
		return;
	regex = *regexp;
	if (regex == NULL)
		return;
	CHECK_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	if (regex->vcl_name != NULL)
		free(regex->vcl_name);
	pcre2_code_free(regex->code);
	FREE_OBJ(regex);
}

VCL_BOOL
vmod_regex_match(VRT_CTX, struct vmod_pcre2_regex *regex,
		 struct vmod_priv *mctx_call, struct vmod_priv *gctx_task,
		 VCL_STRING subject, VCL_INT len, VCL_BOOL anchored,
		 VCL_INT match_limit, VCL_INT offset_limit, VCL_BOOL notbol,
		 VCL_BOOL noteol, VCL_BOOL notempty, VCL_BOOL notempty_atstart,
		 VCL_BOOL no_jit, VCL_BOOL no_utf_check, VCL_ENUM partials,
		 VCL_INT recursion_limit)
{
	pcre2_general_context *gctx;
	pcre2_match_data *mdata;
	struct vmod_priv *match_task = NULL;
	struct match_call *match_opts;
	int ret;
	char *msg;
	uintptr_t snap;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);
	AN(gctx_task);
	AN(mctx_call);
	if (subject == NULL)
		subject = "";

	/*
	 * The general context is task-scoped and shared by all objects
	 * from this VMOD in the task. It uses WS_Alloc as the custom
	 * allocator.
	 */
	if (gctx_task->priv == NULL) {
		gctx = pcre2_general_context_create(ws_malloc, ws_free,
						    ctx->ws);
		if (gctx == NULL) {
			VERRNOMEM(ctx, "initializing general context in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		gctx_task->priv = gctx;
		gctx_task->free = NULL;
	}
	else {
		/*
		 * pcre2_general_context is declared as an incomplete
		 * type, so we can't check if its full size is allocated.
		 */
		WS_Assert_Allocated(ctx->ws, gctx_task->priv, 0);
		gctx = gctx_task->priv;
	}

	/* The match context and options are PRIV_CALL-scoped. */
	/* XXX param to decide whether these should be saved for the call */
	if (mctx_call->priv == NULL) {
		pcre2_match_context *mctx;


		CHECK_UINT32_RANGE(match_limit, regex->vcl_name, ".match()", 0);
		CHECK_UINT32_RANGE(offset_limit, regex->vcl_name, ".match()",
				   0);
		CHECK_UINT32_RANGE(recursion_limit, regex->vcl_name, ".match()",
				   0);

		if ((mctx = pcre2_match_context_create(NULL)) == NULL) {
			VERRNOMEM(ctx, "creating match context in %s.match()",
				  regex->vcl_name);
			return 0;
		}
		SET_CTX_PARAM(mctx, match_limit, uint32_t);
		SET_CTX_PARAM(mctx, offset_limit, uint32_t);
		SET_CTX_PARAM(mctx, recursion_limit, uint32_t);

		ALLOC_OBJ(match_opts, VMOD_PCRE2_MATCH_CALL_MAGIC);
		AN(match_opts);
		match_opts->mctx = mctx;
		mctx_call->priv = match_opts;
		mctx_call->free = match_call_free;

		set_opt(&match_opts->options, anchored,	PCRE2_ANCHORED);
		set_opt(&match_opts->options, notbol,	PCRE2_NOTBOL);
		set_opt(&match_opts->options, noteol,	PCRE2_NOTEOL);
		set_opt(&match_opts->options, notempty,	PCRE2_NOTEMPTY);
		set_opt(&match_opts->options, notempty_atstart,
			PCRE2_NOTEMPTY_ATSTART);
		set_opt(&match_opts->options, no_jit,	PCRE2_NO_JIT);
		set_opt(&match_opts->options, no_utf_check, PCRE2_NO_UTF_CHECK);

		if (partials != NULL) {
			VERR(ctx,
			   "Partial matching not implemented yet in %s.match()",
			     regex->vcl_name);
			return 0;
		}
	}
	else
		CAST_OBJ(match_opts, mctx_call->priv,
			 VMOD_PCRE2_MATCH_CALL_MAGIC);

	/* The match data block is task-scoped for this object only. */
	match_task = VRT_priv_task(ctx, regex);
	AN(match_task);
	if (match_task->priv == NULL) {
		mdata = pcre2_match_data_create_from_pattern(regex->code, gctx);
		if (mdata == NULL) {
			VERRNOMEM(ctx, "initializing match data in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		match_task->priv = mdata;
		match_task->free = NULL;
	}
	else {
		/*
		 * As with the general context above, pcre2_match_data is
		 * an incomplete type.
		 */
		WS_Assert_Allocated(ctx->ws, match_task->priv, 0);
		mdata = match_task->priv;
	}

	if (len == 0)
		len = PCRE2_ZERO_TERMINATED;

	/* XXX param for start_offset */
	ret = pcre2_match(regex->code, (PCRE2_SPTR)subject, (PCRE2_SIZE)len,
			  0, match_opts->options, mdata, match_opts->mctx);
	if (ret == PCRE2_ERROR_NOMATCH)
		return 0;
	assert(ret != PCRE2_ERROR_PARTIAL); /* XXX */
	if (ret >= 0)
		return 1;

	snap = WS_Snapshot(ctx->ws);
	if ((msg = WS_Printf(ctx->ws, "in %s.match()", regex->vcl_name))
	    == NULL)
		msg = "";
	report_pcre2_err(ctx, ret, msg, "");
	WS_Reset(ctx->ws, snap);
	return 0;
}

VCL_STRING
vmod_regex_backref(VRT_CTX, struct vmod_pcre2_regex *regex, VCL_INT ref,
		   VCL_STRING fallback)
{
	struct vmod_priv *match_task;
	pcre2_match_data *mdata;
	PCRE2_UCHAR *buf;
	PCRE2_SIZE len;
	int ret;
	char *msg;
	uintptr_t snap;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);
	CHECK_UINT32_RANGE(ref, regex->vcl_name, ".backref()", fallback);

	match_task = VRT_priv_task(ctx, regex);
	if (match_task == NULL) {
		VERR(ctx, "%s.backref(%ld) called without prior match",
		     regex->vcl_name, ref);
		return fallback;
	}
	WS_Assert_Allocated(ctx->ws, match_task->priv, 0);
	mdata = match_task->priv;

	if ((ret = pcre2_substring_get_bynumber(mdata, (uint32_t)ref, &buf,
						&len)) == 0) {
		WS_Assert_Allocated(ctx->ws, buf, len + 1);
		return (VCL_STRING)buf;
	}

	snap = WS_Snapshot(ctx->ws);
	if ((msg = WS_Printf(ctx->ws, "in %s.backref(%ld)", regex->vcl_name,
			     ref)) == NULL)
		msg = "";
	report_pcre2_err(ctx, ret, msg, "");
	WS_Reset(ctx->ws, snap);
	return fallback;
}

/* Functions */

VCL_BOOL
vmod_config_bool(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what, where;
	int ret;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	
	if (strcmp(configs, "JIT") == 0)
		what = PCRE2_CONFIG_JIT;
	else if (strcmp(configs, "STACKRECURSE") == 0)
		what = PCRE2_CONFIG_STACKRECURSE;
	else if (strcmp(configs, "UNICODE") == 0)
		what = PCRE2_CONFIG_UNICODE;
	else
		WRONG("Illegal config enum in config_bool");

	ret = pcre2_config(what, &where);
	assert(ret >= 0);
	return (VCL_BOOL) where;
}

VCL_STRING
vmod_config_str(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what;
	int ret;
	void *buf;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (strcmp(configs, "BSR") == 0)
		what = PCRE2_CONFIG_BSR;
	else if (strcmp(configs, "JITTARGET") == 0)
		what = PCRE2_CONFIG_JITTARGET;
	else if (strcmp(configs, "NEWLINE") == 0)
		what = PCRE2_CONFIG_NEWLINE;
	else if (strcmp(configs, "UNICODE_VERSION") == 0)
		what = PCRE2_CONFIG_UNICODE_VERSION;
	else if (strcmp(configs, "VERSION") == 0)
		what = PCRE2_CONFIG_VERSION;
	else
		WRONG("Illegal config enum in config_str");


	if (what == PCRE2_CONFIG_NEWLINE || what == PCRE2_CONFIG_BSR) {
		uint32_t where;

		ret = pcre2_config(what, &where);
		assert(ret >= 0);

		if (what == PCRE2_CONFIG_NEWLINE)
			switch(where) {
			case PCRE2_NEWLINE_CR:
				return "CR";
			case PCRE2_NEWLINE_LF:
				return "LF";
			case PCRE2_NEWLINE_CRLF:
				return "CRLF";
			case PCRE2_NEWLINE_ANY:
				return "ANY";
			case PCRE2_NEWLINE_ANYCRLF:
				return "ANYCRLF";
			default:
				WRONG("Illegal result for "
				      "PCRE2_CONFIG_NEWLINE");
			}

		assert (what == PCRE2_CONFIG_BSR);
		switch(where) {
		case PCRE2_BSR_UNICODE:
			return "UNICODE";
		case PCRE2_BSR_ANYCRLF:
			return "ANYCRLF";
		default:
			WRONG("Illegal result for PCRE2_CONFIG_BSR");
		}
	}

	ret = pcre2_config(what, NULL);
	if (what == PCRE2_CONFIG_JITTARGET && ret == PCRE2_ERROR_BADOPTION)
		return "JIT not supported";
	assert(ret > 0);
	if ((buf = WS_Alloc(ctx->ws, ret)) == NULL) {
		ERRNOMEM(ctx, "allocating result in pcre2.config_str()");
		return NULL;
	}
	ret = pcre2_config(what, buf);
	assert(ret > 0);
	return (VCL_STRING) buf;
}

VCL_INT
vmod_config_int(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what, where;
	int ret;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (strcmp(configs, "LINKSIZE") == 0)
		what = PCRE2_CONFIG_LINKSIZE;
	else if (strcmp(configs, "MATCHLIMIT") == 0)
		what = PCRE2_CONFIG_MATCHLIMIT;
	else if (strcmp(configs, "PARENSLIMIT") == 0)
		what = PCRE2_CONFIG_PARENSLIMIT;
	else if (strcmp(configs, "RECURSIONLIMIT") == 0)
		what = PCRE2_CONFIG_RECURSIONLIMIT;
	else
		WRONG("Illegal config enum in config_int");

	ret = pcre2_config(what, &where);
	assert(ret >= 0);
	return (VCL_INT) where;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
